# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.fastapis.stable_xl.text2image import (
    StableXLForText2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_xl.image2image import (
    StableXLForImage2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_xl.inpainting import (
    StableXLForImageInpaintingFastAPIPipeline,
)
