# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.fastapis.stable.text2image import StableForText2ImageFastAPIPipeline
from unitorch.cli.fastapis.stable.image2image import StableForImage2ImageFastAPIPipeline
from unitorch.cli.fastapis.stable.inpainting import (
    StableForImageInpaintingFastAPIPipeline,
)
from unitorch.cli.fastapis.stable.resolution import (
    StableForImageResolutionFastAPIPipeline,
)
from unitorch.cli.fastapis.stable.image2video import StableForImage2VideoFastAPIPipeline

import unitorch.cli.fastapis.stable.interrogator
