# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import unitorch.cli.fastapis.info
import unitorch.cli.fastapis.bria
import unitorch.cli.fastapis.llava
import unitorch.cli.fastapis.stable
import unitorch.cli.fastapis.controlnet
import unitorch.cli.fastapis.stable_xl
import unitorch.cli.fastapis.controlnet_xl
import unitorch.cli.fastapis.stable_3
import unitorch.cli.fastapis.controlnet_3
import unitorch.cli.fastapis.stable_flux
import unitorch.cli.fastapis.controlnet_flux
import unitorch.cli.fastapis.wan
