
@font-face {
  font-family: 'ui-sans-serif';
  src: url('static/fonts/ui-sans-serif/ui-sans-serif-Regular.woff2') format('woff2');
  font-weight: Regular;
  font-style: normal;
}


@font-face {
  font-family: 'ui-sans-serif';
  src: url('static/fonts/ui-sans-serif/ui-sans-serif-Bold.woff2') format('woff2');
  font-weight: Bold;
  font-style: normal;
}


@font-face {
  font-family: 'system-ui';
  src: url('static/fonts/system-ui/system-ui-Regular.woff2') format('woff2');
  font-weight: Regular;
  font-style: normal;
}


@font-face {
  font-family: 'system-ui';
  src: url('static/fonts/system-ui/system-ui-Bold.woff2') format('woff2');
  font-weight: Bold;
  font-style: normal;
}


@font-face {
  font-family: 'sans-serif';
  src: url('static/fonts/sans-serif/sans-serif-Regular.woff2') format('woff2');
  font-weight: Regular;
  font-style: normal;
}


@font-face {
  font-family: 'sans-serif';
  src: url('static/fonts/sans-serif/sans-serif-Bold.woff2') format('woff2');
  font-weight: Bold;
  font-style: normal;
}


@font-face {
  font-family: 'IBM Plex Mono';
  src: url('static/fonts/IBMPlexMono/IBMPlexMono-Regular.woff2') format('woff2');
  font-weight: Regular;
  font-style: normal;
}


@font-face {
  font-family: 'IBM Plex Mono';
  src: url('static/fonts/IBMPlexMono/IBMPlexMono-Bold.woff2') format('woff2');
  font-weight: Bold;
  font-style: normal;
}


@font-face {
  font-family: 'ui-monospace';
  src: url('static/fonts/ui-monospace/ui-monospace-Regular.woff2') format('woff2');
  font-weight: Regular;
  font-style: normal;
}


@font-face {
  font-family: 'ui-monospace';
  src: url('static/fonts/ui-monospace/ui-monospace-Bold.woff2') format('woff2');
  font-weight: Bold;
  font-style: normal;
}


@font-face {
  font-family: 'Consolas';
  src: url('static/fonts/Consolas/Consolas-Regular.woff2') format('woff2');
  font-weight: Regular;
  font-style: normal;
}


@font-face {
  font-family: 'Consolas';
  src: url('static/fonts/Consolas/Consolas-Bold.woff2') format('woff2');
  font-weight: Bold;
  font-style: normal;
}


@font-face {
  font-family: 'monospace';
  src: url('static/fonts/monospace/monospace-Regular.woff2') format('woff2');
  font-weight: Regular;
  font-style: normal;
}


@font-face {
  font-family: 'monospace';
  src: url('static/fonts/monospace/monospace-Bold.woff2') format('woff2');
  font-weight: Bold;
  font-style: normal;
}

:root {
  --name: sudeepshouche/minimalist;
  --background-fill-primary: white;
  --background-fill-secondary: var(--neutral-50);
  --block-background-fill: var(--background-fill-primary);
  --block-border-color: var(--border-color-primary);
  --block-border-width: 0px;
  --block-info-text-color: var(--body-text-color-subdued);
  --block-info-text-size: var(--text-sm);
  --block-info-text-weight: 400;
  --block-label-background-fill: var(--background-fill-primary);
  --block-label-border-color: var(--border-color-primary);
  --block-label-border-width: 1px;
  --block-label-margin: var(--spacing-md);
  --block-label-padding: var(--spacing-sm) var(--spacing-lg);
  --block-label-radius: calc(var(--radius-lg) - 1px) 0 calc(var(--radius-lg) - 1px) 0;
  --block-label-right-radius: 0 calc(var(--radius-lg) - 1px) 0 calc(var(--radius-lg) - 1px);
  --block-label-text-color: var(--neutral-500);
  --block-label-text-size: var(--text-sm);
  --block-label-text-weight: 500;
  --block-padding: var(--spacing-xl) calc(var(--spacing-xl) + 1px);
  --block-radius: var(--radius-lg);
  --block-shadow: none;
  --block-title-background-fill: none;
  --block-title-border-color: none;
  --block-title-border-width: 0px;
  --block-title-padding: var(--block-label-padding);
  --block-title-radius: none;
  --block-title-text-color: var(--neutral-500);
  --block-title-text-size: var(--text-md);
  --block-title-text-weight: 500;
  --body-background-fill: var(--background-fill-primary);
  --body-text-color: var(--neutral-800);
  --body-text-color-subdued: var(--neutral-400);
  --body-text-size: var(--text-md);
  --body-text-weight: 400;
  --border-color-accent: var(--primary-300);
  --border-color-primary: var(--neutral-200);
  --button-border-width: var(--input-border-width);
  --button-cancel-background-fill: var(--button-secondary-background-fill);
  --button-cancel-background-fill-hover: var(--button-cancel-background-fill);
  --button-cancel-border-color: var(--button-secondary-border-color);
  --button-cancel-border-color-hover: var(--button-cancel-border-color);
  --button-cancel-text-color: var(--button-secondary-text-color);
  --button-cancel-text-color-hover: var(--button-cancel-text-color);
  --button-large-padding: var(--spacing-lg) calc(1.5 * var(--spacing-lg));
  --button-large-radius: var(--radius-lg);
  --button-large-text-size: var(--text-lg);
  --button-large-text-weight: 600;
  --button-primary-background-fill: var(--primary-200);
  --button-primary-background-fill-hover: var(--button-primary-background-fill);
  --button-primary-border-color: var(--primary-200);
  --button-primary-border-color-hover: var(--button-primary-border-color);
  --button-primary-text-color: var(--primary-600);
  --button-primary-text-color-hover: var(--button-primary-text-color);
  --button-secondary-background-fill: var(--neutral-200);
  --button-secondary-background-fill-hover: var(--button-secondary-background-fill);
  --button-secondary-border-color: var(--neutral-200);
  --button-secondary-border-color-hover: var(--button-secondary-border-color);
  --button-secondary-text-color: var(--neutral-700);
  --button-secondary-text-color-hover: var(--button-secondary-text-color);
  --button-shadow: none;
  --button-shadow-active: none;
  --button-shadow-hover: none;
  --button-small-padding: var(--spacing-sm) calc(1.5 * var(--spacing-sm));
  --button-small-radius: var(--radius-lg);
  --button-small-text-size: var(--text-md);
  --button-small-text-weight: 400;
  --button-transition: background-color 0.2s ease;
  --checkbox-background-color: var(--background-fill-primary);
  --checkbox-background-color-focus: var(--checkbox-background-color);
  --checkbox-background-color-hover: var(--checkbox-background-color);
  --checkbox-background-color-selected: var(--secondary-600);
  --checkbox-border-color: var(--neutral-300);
  --checkbox-border-color-focus: var(--secondary-500);
  --checkbox-border-color-hover: var(--neutral-300);
  --checkbox-border-color-selected: var(--secondary-600);
  --checkbox-border-radius: var(--radius-sm);
  --checkbox-border-width: var(--input-border-width);
  --checkbox-check: url("data:image/svg+xml,%3csvg viewBox=\'0 0 16 16\' fill=\'white\' xmlns=\'http://www.w3.org/2000/svg\'%3e%3cpath d=\'M12.207 4.793a1 1 0 010 1.414l-5 5a1 1 0 01-1.414 0l-2-2a1 1 0 011.414-1.414L6.5 9.086l4.293-4.293a1 1 0 011.414 0z\'/%3e%3c/svg%3e");
  --checkbox-label-background-fill: var(--button-secondary-background-fill);
  --checkbox-label-background-fill-hover: var(--button-secondary-background-fill-hover);
  --checkbox-label-background-fill-selected: var(--checkbox-label-background-fill);
  --checkbox-label-border-color: var(--border-color-primary);
  --checkbox-label-border-color-hover: var(--checkbox-label-border-color);
  --checkbox-label-border-width: var(--input-border-width);
  --checkbox-label-gap: var(--spacing-lg);
  --checkbox-label-padding: var(--spacing-md) calc(1.5 * var(--spacing-md));
  --checkbox-label-shadow: none;
  --checkbox-label-text-color: var(--body-text-color);
  --checkbox-label-text-color-selected: var(--checkbox-label-text-color);
  --checkbox-label-text-size: var(--text-md);
  --checkbox-label-text-weight: 400;
  --checkbox-shadow: var(--input-shadow);
  --color-accent: var(--primary-500);
  --color-accent-soft: var(--primary-50);
  --container-radius: var(--radius-lg);
  --embed-radius: var(--radius-lg);
  --error-background-fill: #fee2e2;
  --error-border-color: #fecaca;
  --error-border-width: 1px;
  --error-text-color: #ef4444;
  --font: \'Source Sans Pro\', \'ui-sans-serif\', \'system-ui\', sans-serif;
  --font-mono: \'IBM Plex Mono\', \'ui-monospace\', \'Consolas\', monospace;
  --form-gap-width: 0px;
  --input-background-fill: var(--neutral-100);
  --input-background-fill-focus: var(--input-background-fill);
  --input-background-fill-hover: var(--input-background-fill);
  --input-border-color: var(--border-color-primary);
  --input-border-color-focus: var(--secondary-300);
  --input-border-color-hover: var(--input-border-color);
  --input-border-width: 0px;
  --input-padding: var(--spacing-xl);
  --input-placeholder-color: var(--neutral-400);
  --input-radius: var(--radius-lg);
  --input-shadow: none;
  --input-shadow-focus: var(--input-shadow);
  --input-text-size: var(--text-md);
  --input-text-weight: 400;
  --layout-gap: var(--spacing-xxl);
  --link-text-color: var(--secondary-600);
  --link-text-color-active: var(--secondary-600);
  --link-text-color-hover: var(--secondary-700);
  --link-text-color-visited: var(--secondary-500);
  --loader-color: var(--color-accent);
  --neutral-100: #f3f4f6;
  --neutral-200: #e5e7eb;
  --neutral-300: #d1d5db;
  --neutral-400: #9ca3af;
  --neutral-50: #f9fafb;
  --neutral-500: #6b7280;
  --neutral-600: #4b5563;
  --neutral-700: #374151;
  --neutral-800: #1f2937;
  --neutral-900: #111827;
  --neutral-950: #0b0f19;
  --panel-background-fill: var(--background-fill-secondary);
  --panel-border-color: var(--border-color-primary);
  --panel-border-width: 0;
  --primary-100: #dbeafe;
  --primary-200: #bfdbfe;
  --primary-300: #93c5fd;
  --primary-400: #60a5fa;
  --primary-50: #eff6ff;
  --primary-500: #3b82f6;
  --primary-600: #2563eb;
  --primary-700: #1d4ed8;
  --primary-800: #1e40af;
  --primary-900: #1e3a8a;
  --primary-950: #1d3660;
  --prose-header-text-weight: 600;
  --prose-text-size: var(--text-md);
  --prose-text-weight: 400;
  --radio-circle: url("data:image/svg+xml,%3csvg viewBox=\'0 0 16 16\' fill=\'white\' xmlns=\'http://www.w3.org/2000/svg\'%3e%3ccircle cx=\'8\' cy=\'8\' r=\'3\'/%3e%3c/svg%3e");
  --radius-lg: 8px;
  --radius-md: 6px;
  --radius-sm: 4px;
  --radius-xl: 12px;
  --radius-xs: 2px;
  --radius-xxl: 22px;
  --radius-xxs: 1px;
  --secondary-100: #dbeafe;
  --secondary-200: #bfdbfe;
  --secondary-300: #93c5fd;
  --secondary-400: #60a5fa;
  --secondary-50: #eff6ff;
  --secondary-500: #3b82f6;
  --secondary-600: #2563eb;
  --secondary-700: #1d4ed8;
  --secondary-800: #1e40af;
  --secondary-900: #1e3a8a;
  --secondary-950: #1d3660;
  --section-header-text-size: var(--text-md);
  --section-header-text-weight: 400;
  --shadow-drop: rgba(0,0,0,0.05) 0px 1px 2px 0px;
  --shadow-drop-lg: 0 1px 3px 0 rgb(0 0 0 / 0.1), 0 1px 2px -1px rgb(0 0 0 / 0.1);
  --shadow-inset: rgba(0,0,0,0.05) 0px 2px 4px 0px inset;
  --shadow-spread: 3px;
  --slider-color: var(--body-text-color-subdued);
  --spacing-lg: 8px;
  --spacing-md: 6px;
  --spacing-sm: 4px;
  --spacing-xl: 10px;
  --spacing-xs: 2px;
  --spacing-xxl: 16px;
  --spacing-xxs: 1px;
  --stat-background-fill: var(--primary-300);
  --table-border-color: var(--neutral-300);
  --table-even-background-fill: white;
  --table-odd-background-fill: var(--neutral-50);
  --table-radius: var(--radius-lg);
  --table-row-focus: var(--color-accent-soft);
  --text-lg: 16px;
  --text-md: 14px;
  --text-sm: 12px;
  --text-xl: 22px;
  --text-xs: 10px;
  --text-xxl: 26px;
  --text-xxs: 9px;
  --block-label-shadow: var(--block-shadow);
  --border-color-accent-subdued: var(--border-color-accent);
  --code-background-fill: var(--neutral-100);
  --error-icon-color: #b91c1c;
}

.dark {
  --background-fill-primary: var(--neutral-950);
  --background-fill-secondary: var(--neutral-900);
  --block-background-fill: var(--neutral-800);
  --block-border-color: var(--border-color-primary);
  --block-info-text-color: var(--body-text-color-subdued);
  --block-label-background-fill: var(--background-fill-secondary);
  --block-label-border-color: var(--border-color-primary);
  --block-label-text-color: var(--neutral-200);
  --block-title-text-color: var(--neutral-200);
  --body-background-fill: var(--background-fill-primary);
  --body-text-color: var(--neutral-100);
  --body-text-color-subdued: var(--neutral-400);
  --border-color-accent: var(--neutral-600);
  --border-color-primary: var(--neutral-700);
  --button-border-width: var(--input-border-width);
  --button-cancel-background-fill: var(--button-secondary-background-fill);
  --button-cancel-background-fill-hover: var(--button-cancel-background-fill);
  --button-cancel-border-color: var(--button-secondary-border-color);
  --button-cancel-border-color-hover: var(--button-cancel-border-color);
  --button-cancel-text-color: var(--button-secondary-text-color);
  --button-cancel-text-color-hover: var(--button-cancel-text-color);
  --button-primary-background-fill: var(--primary-700);
  --button-primary-background-fill-hover: var(--button-primary-background-fill);
  --button-primary-border-color: var(--primary-600);
  --button-primary-border-color-hover: var(--button-primary-border-color);
  --button-primary-text-color: white;
  --button-primary-text-color-hover: var(--button-primary-text-color);
  --button-secondary-background-fill: var(--neutral-600);
  --button-secondary-background-fill-hover: var(--button-secondary-background-fill);
  --button-secondary-border-color: var(--neutral-600);
  --button-secondary-border-color-hover: var(--button-secondary-border-color);
  --button-secondary-text-color: white;
  --button-secondary-text-color-hover: var(--button-secondary-text-color);
  --checkbox-background-color: var(--neutral-800);
  --checkbox-background-color-focus: var(--checkbox-background-color);
  --checkbox-background-color-hover: var(--checkbox-background-color);
  --checkbox-background-color-selected: var(--secondary-600);
  --checkbox-border-color: var(--neutral-700);
  --checkbox-border-color-focus: var(--secondary-500);
  --checkbox-border-color-hover: var(--neutral-600);
  --checkbox-border-color-selected: var(--secondary-600);
  --checkbox-border-width: var(--input-border-width);
  --checkbox-label-background-fill: var(--button-secondary-background-fill);
  --checkbox-label-background-fill-hover: var(--button-secondary-background-fill-hover);
  --checkbox-label-background-fill-selected: var(--checkbox-label-background-fill);
  --checkbox-label-border-color: var(--border-color-primary);
  --checkbox-label-border-color-hover: var(--checkbox-label-border-color);
  --checkbox-label-border-width: var(--input-border-width);
  --checkbox-label-text-color: var(--body-text-color);
  --checkbox-label-text-color-selected: var(--checkbox-label-text-color);
  --color-accent-soft: var(--neutral-700);
  --error-background-fill: var(--background-fill-primary);
  --error-border-color: var(--border-color-primary);
  --error-text-color: #ef4444;
  --input-background-fill: var(--neutral-700);
  --input-background-fill-focus: var(--secondary-600);
  --input-background-fill-hover: var(--input-background-fill);
  --input-border-color: var(--border-color-primary);
  --input-border-color-focus: var(--neutral-700);
  --input-border-color-hover: var(--input-border-color);
  --input-placeholder-color: var(--neutral-500);
  --link-text-color-active: var(--secondary-500);
  --link-text-color: var(--secondary-500);
  --link-text-color-hover: var(--secondary-400);
  --link-text-color-visited: var(--secondary-600);
  --panel-background-fill: var(--background-fill-secondary);
  --panel-border-color: var(--border-color-primary);
  --shadow-spread: 1px;
  --stat-background-fill: var(--primary-500);
  --table-border-color: var(--neutral-700);
  --table-even-background-fill: var(--neutral-950);
  --table-odd-background-fill: var(--neutral-900);
  --table-row-focus: var(--color-accent-soft);
  --block_border_width: None;
  --block_label_border_width: None;
  --block_shadow: None;
  --block_title_background_fill: None;
  --block_title_border_color: None;
  --block_title_border_width: None;
  --panel_border_width: None;
  --border-color-accent-subdued: var(--border-color-accent);
  --code-background-fill: var(--neutral-800);
  --error_border_width: None;
  --error-icon-color: #ef4444;
  --input_border_width: None;
  --input_shadow: None;
  --input_shadow_focus: None;
  --loader_color: None;
  --slider_color: var(--body-text-color-subdued);
  --name: sudeepshouche/minimalist;
  --block-border-width: 0px;
  --block-info-text-size: var(--text-sm);
  --block-info-text-weight: 400;
  --block-label-border-width: 1px;
  --block-label-margin: var(--spacing-md);
  --block-label-padding: var(--spacing-sm) var(--spacing-lg);
  --block-label-radius: calc(var(--radius-lg) - 1px) 0 calc(var(--radius-lg) - 1px) 0;
  --block-label-right-radius: 0 calc(var(--radius-lg) - 1px) 0 calc(var(--radius-lg) - 1px);
  --block-label-text-size: var(--text-sm);
  --block-label-text-weight: 500;
  --block-padding: var(--spacing-xl) calc(var(--spacing-xl) + 1px);
  --block-radius: var(--radius-lg);
  --block-shadow: none;
  --block-title-background-fill: none;
  --block-title-border-color: none;
  --block-title-border-width: 0px;
  --block-title-padding: var(--block-label-padding);
  --block-title-radius: none;
  --block-title-text-size: var(--text-md);
  --block-title-text-weight: 500;
  --body-text-size: var(--text-md);
  --body-text-weight: 400;
  --button-large-padding: var(--spacing-lg) calc(1.5 * var(--spacing-lg));
  --button-large-radius: var(--radius-lg);
  --button-large-text-size: var(--text-lg);
  --button-large-text-weight: 600;
  --button-shadow: none;
  --button-shadow-active: none;
  --button-shadow-hover: none;
  --button-small-padding: var(--spacing-sm) calc(1.5 * var(--spacing-sm));
  --button-small-radius: var(--radius-lg);
  --button-small-text-size: var(--text-md);
  --button-small-text-weight: 400;
  --button-transition: background-color 0.2s ease;
  --checkbox-border-radius: var(--radius-sm);
  --checkbox-check: url("data:image/svg+xml,%3csvg viewBox=\'0 0 16 16\' fill=\'white\' xmlns=\'http://www.w3.org/2000/svg\'%3e%3cpath d=\'M12.207 4.793a1 1 0 010 1.414l-5 5a1 1 0 01-1.414 0l-2-2a1 1 0 011.414-1.414L6.5 9.086l4.293-4.293a1 1 0 011.414 0z\'/%3e%3c/svg%3e");
  --checkbox-label-gap: var(--spacing-lg);
  --checkbox-label-padding: var(--spacing-md) calc(1.5 * var(--spacing-md));
  --checkbox-label-shadow: none;
  --checkbox-label-text-size: var(--text-md);
  --checkbox-label-text-weight: 400;
  --checkbox-shadow: var(--input-shadow);
  --color-accent: var(--primary-500);
  --container-radius: var(--radius-lg);
  --embed-radius: var(--radius-lg);
  --error-border-width: 1px;
  --font: \'Source Sans Pro\', \'ui-sans-serif\', \'system-ui\', sans-serif;
  --font-mono: \'IBM Plex Mono\', \'ui-monospace\', \'Consolas\', monospace;
  --form-gap-width: 0px;
  --input-border-width: 0px;
  --input-padding: var(--spacing-xl);
  --input-radius: var(--radius-lg);
  --input-shadow: none;
  --input-shadow-focus: var(--input-shadow);
  --input-text-size: var(--text-md);
  --input-text-weight: 400;
  --layout-gap: var(--spacing-xxl);
  --loader-color: var(--color-accent);
  --neutral-100: #f3f4f6;
  --neutral-200: #e5e7eb;
  --neutral-300: #d1d5db;
  --neutral-400: #9ca3af;
  --neutral-50: #f9fafb;
  --neutral-500: #6b7280;
  --neutral-600: #4b5563;
  --neutral-700: #374151;
  --neutral-800: #1f2937;
  --neutral-900: #111827;
  --neutral-950: #0b0f19;
  --panel-border-width: 0;
  --primary-100: #dbeafe;
  --primary-200: #bfdbfe;
  --primary-300: #93c5fd;
  --primary-400: #60a5fa;
  --primary-50: #eff6ff;
  --primary-500: #3b82f6;
  --primary-600: #2563eb;
  --primary-700: #1d4ed8;
  --primary-800: #1e40af;
  --primary-900: #1e3a8a;
  --primary-950: #1d3660;
  --prose-header-text-weight: 600;
  --prose-text-size: var(--text-md);
  --prose-text-weight: 400;
  --radio-circle: url("data:image/svg+xml,%3csvg viewBox=\'0 0 16 16\' fill=\'white\' xmlns=\'http://www.w3.org/2000/svg\'%3e%3ccircle cx=\'8\' cy=\'8\' r=\'3\'/%3e%3c/svg%3e");
  --radius-lg: 8px;
  --radius-md: 6px;
  --radius-sm: 4px;
  --radius-xl: 12px;
  --radius-xs: 2px;
  --radius-xxl: 22px;
  --radius-xxs: 1px;
  --secondary-100: #dbeafe;
  --secondary-200: #bfdbfe;
  --secondary-300: #93c5fd;
  --secondary-400: #60a5fa;
  --secondary-50: #eff6ff;
  --secondary-500: #3b82f6;
  --secondary-600: #2563eb;
  --secondary-700: #1d4ed8;
  --secondary-800: #1e40af;
  --secondary-900: #1e3a8a;
  --secondary-950: #1d3660;
  --section-header-text-size: var(--text-md);
  --section-header-text-weight: 400;
  --shadow-drop: rgba(0,0,0,0.05) 0px 1px 2px 0px;
  --shadow-drop-lg: 0 1px 3px 0 rgb(0 0 0 / 0.1), 0 1px 2px -1px rgb(0 0 0 / 0.1);
  --shadow-inset: rgba(0,0,0,0.05) 0px 2px 4px 0px inset;
  --spacing-lg: 8px;
  --spacing-md: 6px;
  --spacing-sm: 4px;
  --spacing-xl: 10px;
  --spacing-xs: 2px;
  --spacing-xxl: 16px;
  --spacing-xxs: 1px;
  --table-radius: var(--radius-lg);
  --text-lg: 16px;
  --text-md: 14px;
  --text-sm: 12px;
  --text-xl: 22px;
  --text-xs: 10px;
  --text-xxl: 26px;
  --text-xxs: 9px;
  --block-label-shadow: var(--block-shadow);
}

div.tabitem {
  padding: 10px 0px 0px 0px; 
  border: 0px;
}

@media (min-width: 1536px)
{
    .gradio-container {
        min-width: 80% !important;
    }
}

img {
  max-height: 500px;
}

video {
  max-height: 500px;
}

.ut-bg-transparent {
  background-color: transparent !important;
}

.ut-a-none {
  text-decoration: none;
}

.ut-0-margin-padding {
  margin: 0;
  padding: 0;
}

canvas {
  max-width: 100% !important;
  max-height: 100% !important;
}

@media (max-width: 1536px)
{
  .ut-place-holder-block {
    display: none;
  }
}
table {
  width: 100%;
}

footer {
  visibility: hidden;
}