# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.fastapis.controlnet_xl.text2image import (
    ControlNetXLForText2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.controlnet_xl.image2image import (
    ControlNetXLForImage2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.controlnet_xl.inpainting import (
    ControlNetXLForImageInpaintingFastAPIPipeline,
)
