# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.xpegasus.modeling import XPegasusForGeneration
from unitorch.models.xpegasus.processing import (
    XPegasusProcessor,
    get_xpegasus_tokenizer,
)
