# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.bert.modeling import BertForClassification
from unitorch.models.bert.processing import BertProcessor, get_bert_tokenizer
