# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.fastapis.stable_3.text2image import (
    Stable3ForText2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_3.image2image import (
    Stable3ForImage2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_3.inpainting import (
    Stable3ForImageInpaintingFastAPIPipeline,
)
