# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.dinov2.modeling import DinoV2ForImageClassification
from unitorch.models.dinov2.processing import DinoV2Processor
