# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.fastapis.stable_flux.text2image import (
    StableFluxForText2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_flux.image2image import (
    StableFluxForImage2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_flux.image_control import (
    StableFluxForImageControlGenerationFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_flux.image_redux import (
    StableFluxForImageReduxGenerationFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_flux.inpainting import (
    StableFluxForImageInpaintingFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_flux.redux_inpainting import (
    StableFluxForReduxInpaintingFastAPIPipeline,
)
